#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-

import sys
sys.dont_write_bytecode = True

import os
import re
import fcntl
from datetime import datetime

SUB_DIR_FMT = "{:04d}{:02d}"
BASE_NAME_FMT = "{:04d}{:02d}{:02d}"
LOG_DATA_FMT = "{:04d}/{:02d}/{:02d} {:02d}:{:02d}:{:02d},{}"

LOG_SUFFIX_RE = r".*_([0-9]+)\.log$"

LOG_INFO_FMT = "INFO,{}"
LOG_WARNING_FMT = "WARNING,{}"
LOG_ERROR_FMT = "ERROR,{}"


class Log(object):
    """logger class for logging about meimeikun file copy."""

    LOG_FILE_MODE = "ab"
    DEFAULT_LOG_MAX_BYTES = 10 * 1024 * 1024

    def __init__(self, log_dir,
                 log_max_bytes=DEFAULT_LOG_MAX_BYTES,
                 line_sep=os.linesep):
        """Constructor."""

        self.log_dir = log_dir
        self.log_max_bytes = log_max_bytes
        self.line_sep = line_sep
        self.opened_file = None

    def __del__(self):
        """Destructor."""

        self.close()

    def __enter__(self):
        """Enter context for logging."""

        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit context for logging."""

        self.close()

    def open(self):
        """Open and Lock log file whose name is based on "start_epoch"."""

        if self.opened_file is not None:
            self.close()

        now = datetime.now()
        sub_dir = SUB_DIR_FMT.format(now.year, now.month)
        base_name = BASE_NAME_FMT.format(now.year, now.month, now.day)
        log_dir = os.path.join(self.log_dir, sub_dir)
        log_file = self._get_log_file_path(log_dir, base_name)

        self.opened_file = open(log_file, self.LOG_FILE_MODE)
        self._lock()
        os.chmod(log_file, 0666)

        return self

    def close(self, condition=None):
        """Close and Unlock log file if it is opened."""
        if self.opened_file is not None:
            self._unlock()
            self.opened_file.close()
            self.opened_file = None

    def write_info(self, data):
        self._write(LOG_INFO_FMT.format(data))

    def write_warning(self, data):
        self._write(LOG_WARNING_FMT.format(data))

    def write_error(self, data):
        self._write(LOG_ERROR_FMT.format(data))

    def _write(self, data):
        if self.opened_file is not None:
            now = datetime.now()
            self.opened_file.write(
                LOG_DATA_FMT.format(now.year, now.month, now.day,
                                    now.hour, now.minute, now.second, data)
                + self.line_sep)

    def _is_size_exceed(self, file_path):
        return os.path.getsize(file_path) > self.log_max_bytes

    def _get_log_file_path(self, log_dir, base_name):
        file_name = self._get_current_log_file_name(log_dir, base_name)
        if file_name is None:
            file_name = self._make_log_file_name(base_name)
            if not os.path.exists(log_dir):
                self.__makedirs_in_mode(log_dir, 0o777)
        else:
            if self._is_size_exceed(os.path.join(log_dir, file_name)):
                file_name = self._make_log_file_name_with_suffix(file_name)

        return os.path.join(log_dir, file_name)

    def _get_suffix_of_log_file(self, file_name):
        match = re.match(LOG_SUFFIX_RE, file_name)
        if match is None:
            return 0
        else:
            return int(match.groups()[0])

    def _make_log_file_name_with_suffix(self, file_name):
        match = re.match(LOG_SUFFIX_RE, file_name)
        if match is None:
            return re.sub(r"\.log$", "_2.log", file_name)
        else:
            suffix = int(match.groups()[0]) + 1
            return re.sub(r"_([0-9]+)\.log$",
                          "_{}.log".format(suffix), file_name)

    def _make_log_file_name(self, base_name):
        return "{}.log".format(base_name)

    def _get_current_log_file_name(self, log_dir, base_name):
        if not os.path.exists(log_dir):
            return None

        log_list = []
        LOG_FILENAME_RE = r"{}(_[0-9]+)?\.log$".format(base_name)
        for log_file in os.listdir(log_dir):
            if re.match(LOG_FILENAME_RE, log_file) is None:
                continue

            suffix = self._get_suffix_of_log_file(log_file)
            log_list.append((suffix, log_file))

        if not log_list:
            return None

        log_list.sort(reverse=True, key=lambda x: x[0])
        return log_list[0][1]

    def __makedirs_in_mode(self, dir_path, mode):
        if os.path.isdir(dir_path):
            return

        head, tail = os.path.split(dir_path)
        self.__makedirs_in_mode(head, mode)

        os.mkdir(dir_path)
        os.chmod(dir_path, mode)

    def _lock(self):
        fcntl.flock(self.opened_file.fileno(), fcntl.LOCK_EX)

    def _unlock(self):
        fcntl.flock(self.opened_file.fileno(), fcntl.LOCK_UN)
